from pyaxm.client import Client
import sys
import pandas as pd

def list_devices():
    client = Client()
    devices = client.list_devices()
    df = pd.DataFrame(devices)
    df.to_csv(sys.stdout, index=False)

def query_device():
    client = Client()
    if len(sys.argv) < 3:
        print("Usage: pyaxm-cli device <device_id>")
        exit(1)
    device_id = sys.argv[2]
    device = client.get_device(device_id)
    df = pd.DataFrame([device])
    df.to_csv(sys.stdout, index=False)

def list_mdm_servers():
    client = Client()
    servers = client.list_mdm_servers()
    df = pd.DataFrame(servers)
    df.to_csv(sys.stdout, index=False)

def list_devices_in_mdm_server():
    if len(sys.argv) < 3:
        print("Usage: pyaxm-cli mdm_server <server_id>")
        print("You can get the server_id from the 'mdm_servers' command.")
        exit(1)
    server_id = sys.argv[2]
    client = Client()
    devices = client.list_devices_in_mdm_server(server_id)
    df = pd.DataFrame(devices)
    df.to_csv(sys.stdout, index=False)

def main():
    if not len(sys.argv) > 1:
        print("Usage: pyaxm-cli <command> [<args>]")
        print("Available commands: devices device mdm_servers mdm_server")
        exit(1)

    match sys.argv[1]:
        case "devices":
            list_devices()
        case "device":
            query_device()
        case "mdm_servers":
            list_mdm_servers()
        case "mdm_server":
            list_devices_in_mdm_server()
        case _:
            print("Invalid command.")
            print("Available commands: devices device mdm_servers mdm_server")
            exit(1)

if __name__ == "__main__":
    main()




import datetime as dt
from pyaxm.models import OrgDevice
from pyaxm import abm_requests
import datetime as dt
import Cryptodome.PublicKey.ECC as ECC
from authlib.jose import jwt
import uuid
import os
import json
import time

ABM_CLIENT_ID = os.environ['AXM_CLIENT_ID']
ABM_KEY_ID = os.environ['AXM_KEY_ID']
ABM_FOLDER = os.path.join(os.path.expanduser('~'), '.config', 'pyaxm')
KEY_PATH = os.path.join(ABM_FOLDER, 'key.pem')
TOKEN_PATH = os.path.join(ABM_FOLDER, 'token.json')

class AccessToken():
    def __init__(self):
        self.value = None
        self.expires_at = None
        self.assertion = None

        # Try reading from cache
        try:
            with open(TOKEN_PATH, 'r') as f:
                cache = json.load(f)
            
            if cache['expires_at'] > time.time():
                self.value = cache['access_token']
                self.expires_at = dt.datetime.fromtimestamp(cache['expires_at'], dt.timezone.utc)  # Convert timestamp back to datetime
                return
        except (FileNotFoundError, json.JSONDecodeError, KeyError):
            pass
        
        # Get new token
        self.generate_assertion()
        self.generate_token()

        # Write to cache
        cache_data = {
            'access_token': self.value,
            'expires_at': (self.expires_at).timestamp()  # Store as a timestamp
        }

        with open(TOKEN_PATH, 'w') as f:
            json.dump(cache_data, f)
    
    def generate_assertion(self):
        issued_at = int(dt.datetime.now(dt.timezone.utc).timestamp())
        expires_at = issued_at + 60 # expires in 1 minute

        headers = {
            "alg": "ES256",
            "kid": ABM_KEY_ID
        }

        payload = {
            "sub": ABM_CLIENT_ID,
            "aud": 'https://account.apple.com/auth/oauth2/v2/token',
            "iat": issued_at,
            "exp": expires_at,
            "jti": str(uuid.uuid4()),
            "iss": ABM_CLIENT_ID
        }

        with open(KEY_PATH, 'rt') as f:
            private_key = ECC.import_key(f.read())

        self.assertion = jwt.encode(
            header=headers,
            payload=payload,
            key=private_key.export_key(format='PEM')
        ).decode("utf-8")

    def generate_token(self):     
        data = {
            "grant_type": "client_credentials",
            "client_id": ABM_CLIENT_ID,
            "client_assertion_type": 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            "client_assertion": self.assertion,
            "scope": 'business.api'
        }

        token = abm_requests.get_access_token(data)        
        expires_in = token.get('expires_in')
        issued_at = dt.datetime.now(dt.timezone.utc)

        self.expires_at = issued_at + dt.timedelta(seconds=expires_in)
        self.value = token.get('access_token')

class Client:
    def __init__(self):
        self.access_token = AccessToken()
    
    def list_devices(self) -> list[dict]:
        '''Returns a list of devices in the organization.
        '''
        response = abm_requests.list_devices(self.access_token.value)
        devices = [data.attributes.model_dump() for data in response.data]

        while response.links.next:
            next_page = response.links.next
            response = abm_requests.list_devices(self.access_token.value, next=next_page)
            devices.extend([data.attributes.model_dump() for data in response.data])
            
        return devices

    def get_device(self, device_id: str) -> dict:
        '''Returns a device's details as a dict.
        '''
        response = abm_requests.get_device(device_id, self.access_token.value)
        return response.data.attributes.model_dump()

    ## - MDM servers
    def list_mdm_servers(self) -> list[dict]:
        '''Returns a list of MDM servers with their names and IDs.
        TODO: Add pagination as currently it only returns the first page of results.
        '''
        response = abm_requests.list_mdm_servers(self.access_token.value)
        exclude_keys = {
            'data': {
                '__all__': {
                    'relationships',
                    'type'
                },
            },
            'included': True,
            'links': True,
            'meta': True,
        }

        data = response.model_dump(exclude=exclude_keys)
        data = [
            {
                'id': server.id,
                'createdDateTime': server.attributes.createdDateTime,
                'serverName': server.attributes.serverName,
                'serverType': server.attributes.serverType,
                'updatedDateTime': server.attributes.updatedDateTime
            } for server in response.data
        ]

        return data

    ## - list_devices_in_mdm_server
    def list_devices_in_mdm_server(self, server_id: str) -> list[str]:
        '''Returns a list of device IDs (serials) in the specified MDM server.
        '''
        response = abm_requests.list_devices_in_mdm_server(server_id, self.access_token.value)
        include_keys= {
            'data': {
                '__all__': {
                    'id'
                }
            }
        }

        devices = response.model_dump(include=include_keys)
        devices = [device for device in devices['data']]

        while response.links.next:
            next_page = response.links.next
            response = abm_requests.list_devices_in_mdm_server(server_id, self.access_token.value, next=next_page)
            devices_dump = response.model_dump(include=include_keys)
            devices_dump = [device for device in devices_dump['data']]
            devices.extend(devices_dump)

        return devices



from pydantic import BaseModel, ConfigDict
from typing import List, Optional
from enum import Enum

class OrgDeviceActivityType(Enum):
    ASSIGN_DEVICE = "ASSIGN_DEVICE"
    UNASSIGN_DEVICE = "UNASSIGN_DEVICE"

class DocumentLinks(BaseModel):
    self: str # uri-reference to the current document

class Parameter(BaseModel):
    parameter: str

class JsonPointer(BaseModel):
    pointer: str

class ResourceLinks(BaseModel):
    self: Optional[str] # uri-reference. Link to the resource

class RelationshipLinks(BaseModel):
    include: Optional[str] = None # is this really a field? #### TODO: confirm
    related: Optional[str] = None # uri-reference
    self: Optional[str] # uri-reference

class PagedDocumentLinks(BaseModel):
    first: Optional[str] = None # uri-reference
    next: Optional[str] = None # uri-reference
    self: str # uri-reference

# OrgDevice
class OrgDevice(BaseModel):
    class Attributes(BaseModel):
        addedToOrgDateTime: Optional[str] # ISO 8601 date-time
        color: Optional[str]
        deviceCapacity: Optional[str]
        deviceModel: Optional[str]
        eid: Optional[str]
        imei: Optional[List[str]]
        meid: Optional[List[str]]
        orderDateTime: Optional[str] # ISO 8601 date-time
        orderNumber: Optional[str]
        partNumber: Optional[str]
        productFamily: Optional[str]
        productType: Optional[str]
        purchaseSourceType: Optional[str]
        purchaseSourceId: Optional[str]
        serialNumber: Optional[str]
        status: Optional[str]
        updatedDateTime: Optional[str] # ISO 8601 date-time
    
    class Relationships(BaseModel):
        class AssignedServer(BaseModel):
            links: Optional[RelationshipLinks]

        assignedServer: Optional[AssignedServer]

    attributes: Optional[Attributes]
    id: str
    links: Optional[ResourceLinks]
    relationships: Optional[Relationships]
    type: str

class OrgDeviceAssignedServerLinkageResponse(BaseModel):
    class Data(BaseModel):
        id: str
        type: str

    data: Data
    links: DocumentLinks

class OrgDeviceActivity(BaseModel):
    class Attributes(BaseModel):
        createdDateTime: Optional[str] # ISO 8601 date-time
        status: Optional[str]
        subStatus: Optional[str]
        completedDateTime: Optional[str] # ISO 8601 date-time
        downloadUrl: Optional[str]

    attributes: Optional[Attributes]
    id: str
    links: Optional[ResourceLinks]
    type: str

class OrgDeviceActivityCreateRequest(BaseModel):
    class Data(BaseModel):
        class Attributes(BaseModel):
            activityType: OrgDeviceActivityType
        
        class Relationships(BaseModel):
            class Devices(BaseModel):
                class Data(BaseModel):
                    id: str
                    type: str
                
                data: List[Data]
            
            class MdmServer(BaseModel):
                class Data(BaseModel):
                    id: str
                    type: str

                data: Data

            devices: Devices
            mdmServer: MdmServer

        attributes: Attributes
        relationships: Relationships
        type: str
    data: Data

class PagingInformation(BaseModel):
    class Paging(BaseModel):
        limit: int
        nextCursor: Optional[str] = None # also weird not being passed
        total: Optional[int] = None # also not being passed

    paging: Paging

class MdmServer(BaseModel):
    class Attributes(BaseModel):
        createdDateTime: Optional[str] # ISO 8601 date-time
        serverName: Optional[str]
        serverType: Optional[str]
        updatedDateTime: Optional[str] # ISO 8601 date-time
    
    class Relationships(BaseModel):
        class Devices(BaseModel):
            class Data(BaseModel):
                id: str
                type: str

            data: Optional[List[Data]] = None # weird also not being returned
            links: Optional[RelationshipLinks]
            meta: Optional[PagingInformation] = None # also weird

        devices: Optional[Devices]

    attributes: Optional[Attributes]
    id: str
    relationships: Optional[Relationships]
    type: str

class MdmServerLinkageResponse(BaseModel):
    class Data(BaseModel):
        id: str
        type: str

    data: Data
    links: PagedDocumentLinks
    meta: Optional[PagingInformation]

class OrgDeviceActivityResponse(BaseModel):
    data: OrgDeviceActivity
    links: DocumentLinks

class MdmServersResponse(BaseModel):
    data: List[MdmServer]
    included: Optional[List[OrgDevice]] = None # weird not being returned
    links: PagedDocumentLinks
    meta: Optional[PagingInformation]

class MdmServerResponse(BaseModel):
    data: MdmServer
    included: Optional[List[OrgDevice]]
    links: DocumentLinks

class OrgDevicesResponse(BaseModel):
    data: List[OrgDevice]
    links: PagedDocumentLinks
    meta: Optional[PagingInformation]

class OrgDeviceResponse(BaseModel):
    data: OrgDevice
    links: DocumentLinks

class ErrorLinks(BaseModel):
    class Associated(BaseModel):
        class Meta(BaseModel):
            source: Optional[str]
        
        href: Optional[str] # uri-reference
        meta: Optional[Meta]
    
    about: Optional[str] # uri-reference
    associated: Optional[str|Associated] # uri-reference or Associated object

class ErrorResponse(BaseModel):
    class Errors(BaseModel):
        class Meta(BaseModel):
            # allows non-specified key/value pairs
            model_config = ConfigDict(extra='allow')

        code: str
        detail: str
        id: Optional[str]
        source: Optional[JsonPointer|Parameter]
        status: str
        title: str
        links: Optional[ErrorLinks]
        meta: Optional[Meta]
    
    errors: Optional[List[Errors]]

class MdmServerDevicesLinkagesResponse(BaseModel):
    class Data(BaseModel):
        id: str
        type: str

    data: List[Data]
    links: PagedDocumentLinks
    meta: Optional[PagingInformation]



import requests
from pyaxm.models import OrgDeviceResponse, MdmServersResponse, MdmServerDevicesLinkagesResponse, OrgDevicesResponse
import time
from functools import wraps

# creating a session to reuse connections
# didn't really improved performance, should revert back 
# to requests without session?
session = requests.Session()

def exponential_backoff(retries=5, backoff_factor=2):
    """
    A decorator for retrying a function with exponential backoff.
    
    :param retries: Number of retry attempts.
    :param backoff_factor: Factor by which the wait time increases.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(retries):
                try:
                    return func(*args, **kwargs)
                except requests.exceptions.RequestException as e:
                    if attempt < retries - 1:
                        wait_time = backoff_factor ** attempt
                        time.sleep(wait_time)
                    else:
                        raise e
        return wrapper
    return decorator

def _auth_headers(access_token: str) -> dict:
    """
    :param access_token: The access token for authentication.
    :return: A dictionary containing the authorization headers.
    """
    return {"Authorization": f"Bearer {access_token}"}

def get_access_token(data: dict) -> dict:
    """
    Generate an access token for Apple Business Manager API.
    :param data: A dictionary containing the necessary parameters for the token request.
    :return: A dictionary containing the access token and other related information.
    """
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Host': 'account.apple.com'
    }

    response = session.post(
        'https://account.apple.com/auth/oauth2/token',
        headers=headers,
        data=data
    )

    response.raise_for_status()
    return response.json()

@exponential_backoff(retries=5, backoff_factor=2)
def list_devices(access_token, next=None) -> OrgDevicesResponse:
    """
    List all organization devices.
    :param access_token: The access token for authentication.
    :param next: Optional; the URL for the next page of results.
    :return: An OrgDevicesResponse object containing the list of devices.
    """
    if next:
        url = next
    else:
        # Default is 100 devices. Documentation says max is 1000 but 
        # currently querying a unspecified number returns status code 500
        # increase this when apple fixes the issue to 1000
        url = 'https://api-business.apple.com/v1/orgDevices'

    response = session.get(url, headers=_auth_headers(access_token))

    if response.status_code == 200:
        return OrgDevicesResponse.model_validate(response.json())
    else:
        response.raise_for_status()

@exponential_backoff(retries=5, backoff_factor=2)
def get_device(device_id, access_token) -> OrgDeviceResponse:
    """
    Retrieve an organization device by its ID.
    
    :param device_id: The ID of the organization device to retrieve.
    :param access_token: The access token for authentication.
    :return: An OrgDeviceResponse object containing the device information.
    """

    url = f'https://api-business.apple.com/v1/orgDevices/{device_id}'
    response = session.get(url, headers=_auth_headers(access_token))
    
    if response.status_code == 200:
        return OrgDeviceResponse.model_validate(response.json())
    else:
        response.raise_for_status()

@exponential_backoff(retries=5, backoff_factor=2)
def list_mdm_servers(access_token) -> MdmServersResponse:
    """
    List all MDM servers.
    
    :param access_token: The access token for authentication.
    :return: An MdmServersResponse object containing the list of MDM servers.
    """
    url = 'https://api-business.apple.com/v1/mdmServers'    
    response = session.get(url, headers=_auth_headers(access_token))
    
    if response.status_code == 200:
        return MdmServersResponse.model_validate(response.json())
    else:
        response.raise_for_status()

@exponential_backoff(retries=5, backoff_factor=2)
def list_devices_in_mdm_server(server_id: str, access_token, next=None) -> MdmServerDevicesLinkagesResponse:
    """
    List devices in a specific MDM server.
    
    :param server_id: The ID of the MDM server.
    :param access_token: The access token for authentication.
    :param next: Optional; the URL for the next page of results.
    :return: An MdmServerResponse object containing the MDM server information.
    """
    if next:
        url = next
    else:
        url = f'https://api-business.apple.com/v1/mdmServers/{server_id}/relationships/devices?limit=1000'

    response = session.get(url, headers=_auth_headers(access_token))

    # ABM has been returning 500, this is a workaround to retry 2 times
    # before raising an error.
    if response.status_code == 200:
        return MdmServerDevicesLinkagesResponse.model_validate(response.json())
    else:
        response.raise_for_status()
